// Dual-sum checksum examples -- Chapter 6 -- unit tests

// Copyright 2024 Philip Koopman
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//    http://www.apache.org/licenses/LICENSE-2.0
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#include "support.h"
#include "chapter6_examples.h"

static const uint32_t initialSeed = SEED;  // Initial sum

// Test generic dual sum 
uint32_t TryDual16(uint8_t dataWord[], uint32_t dwSize, uint32_t modulus)
{ 
  uint32_t resultA = DualSum16_A(dataWord, dwSize, modulus);
  uint32_t resultB = DualSum16_B(dataWord, dwSize, modulus);
  uint32_t resultC = DualSum16_C(dataWord, dwSize, modulus);
  uint32_t resultD = DualSum16_D(dataWord, dwSize, modulus);
  uint32_t resultE = DualSum16_E(dataWord, dwSize, modulus);

  printf("Modulus=0x%02X\n", modulus);
  printf("Result DualSum16_A=\t0x%02X\n", resultA);
  printf("Result DualSum16_B=\t0x%02X\n", resultB);
  printf("Result DualSum16_C=\t0x%02X\n", resultC);
  printf("Result DualSum16_D=\t0x%02X\n", resultD);
  printf("Result DualSum16_E=\t0x%02X\n", resultE);

  // Make sure they all got the same result  
  assert(resultA == resultB);
  assert(resultB == resultC); 
  assert(resultC == resultD); 
  assert(resultC == resultE); 
  return(resultA);
}

// Test Fletcher special code and compare against generic code
uint32_t TryFletcher16(uint8_t dataWord[], uint32_t dwSize)
{ uint32_t modulus = 255;
  uint32_t resultA = TryDual16(dataWord, dwSize, modulus);
  uint32_t resultF = Fletcher16(dataWord, dwSize, modulus);
  
  printf("Result Fletcher16_D=\t0x%02X\n", resultF);

  if ( (resultF & 0xFF00) == 0xFF00) // top byte -0
  { resultF &= 0x00FF ;  // zero top byte
  }
  if ( (resultF & 0x00FF) == 0x00FF) // bottom byte -0
  { resultF &= 0xFF00 ;  // zero bottom byte
  }
  
  // Make sure they all got the same result  
  assert(resultA == resultF);
  return(resultA);
}

// Test DualX checksum
uint32_t TryDualX16(uint8_t dataWord[], uint32_t dwSize)
{ 
  uint32_t modulus = 253;
  uint32_t resultA = DualX(dataWord, dwSize, modulus);

  printf("Modulus=0x%02X\n", modulus);
  printf("Result DualX=\t0x%02X\n", resultA);

  return(resultA);
}

// -------------------------------
// Run all unit tests
int main(void) 
{ 
#ifdef VALIDATE
  const uint32_t dwAlloc = 16; // divisible by 4 and greater than 9
  uint32_t dwSize = 9; 
#else
  const uint32_t dwAlloc = 65000; // divisible by 4 and greater than 9
  uint32_t dwSize = 65000; 
#endif

  uint8_t dataWord[dwAlloc];
  InitializeData(dataWord, dwAlloc); // initialize padding bytes too
  PrintWord(dataWord, dwSize);

  printf(">>>>> Initial Seed = %d\n", initialSeed);

  uint32_t result = 0;

  result = TryFletcher16(dataWord, dwSize);  // Fletcher16 checksum
#ifdef VALIDATE
  if(initialSeed == 0x00) { assert(result == 0xDE1E); } else 
  if(initialSeed == 0x9D) { assert(result == 0x7C46); } else 
  if(initialSeed == 0xFF) { assert(result == 0xDE1E); } else 
  { assert(false); }
#endif    
 
  result = TryDual16(dataWord, dwSize, 251); // Adler16 checksum
#ifdef VALIDATE
  if(initialSeed == 0x00) { assert(result == 0xE242); } else 
  if(initialSeed == 0x9D) { assert(result == 0x8482); } else 
  if(initialSeed == 0xFF) { assert(result == 0xE66A); } else 
  { assert(false); }
#endif  

  result = TryDual16(dataWord, dwSize, 253); // Dual16 mod 253 checksum
#ifdef VALIDATE
  if(initialSeed == 0x00) { assert(result == 0xE030); } else 
  if(initialSeed == 0x9D) { assert(result == 0x8064); } else 
  if(initialSeed == 0xFF) { assert(result == 0xE244); } else 
  { assert(false); }  
#endif   
  
 // DualX requires even number of bytes in data word 
 uint32_t dwSizeEven = dwSize & 0xFFFFFFFE;
 result = TryDualX16(dataWord, dwSizeEven );  // DualX checksum
#ifdef VALIDATE
  if(initialSeed == 0x00) { assert(result == 0x4D1A); } else 
  if(initialSeed == 0x9D) { assert(result == 0xEA34); } else 
  if(initialSeed == 0xFF) { assert(result == 0x4F24); } else 
  { assert(false); }  
#endif   
  
  printf("<<<<<<<<<<<<<<<<<<<<<<<<<<< PASS >>>>>>>>>>>>>>>>>>>>>>>>>>>\n");
  result = EXIT_SUCCESS;
  return((int)result) ;
}
