// Hamming Distance CRC Polynomial Evaluation
// Copyright 2015-2023, Philip Koopman  koopman@cmu.edu
// Creative Commons Attribution-ShareAlike 4.0 International
//       http://creativecommons.org/licenses/by-sa/4.0/
// No warranty express or implied; use at your own risk
// User assumes responsibility for validating suitability for use
// We suggest using the non-optimized version for validation.

// Slow version -- NO lookup table; NO hash table
//    - But, simpler logic and easier to check for accuracy
//    - Simpler bit count
//    - LFSR has had lookup table, hash table, and search routines removed
//    - CRCPoly has had special cases for HD=3, HD=4 and skipping levels removed
//    - walks all bit patterns across data word; nothing fancy
// Really slow, but the idea is to provide a way to validate the fast version better

// usage:  hdlen  < polylist
//         hdlen poly
//         hdlen startHD stopHD  < polylist
//         hdlen poly startHD stopHD                        

// stdin is hex CRC polynomial up to 64 bits in implicit +1 notation
// if startHD and stopHD are not specified all HD lengths are computed from 3 up
// Example inputs:

//      ./hdlen 0x82608edb
// gives entire HD profile of CRC-32, which is:
//   0x82608edb {4294967263,91607,2974,268,171,91,57,34,21,12,10,10,10}

//      ./hdlen 0x82608edb 5 7
// gives HD=5, 6, and 7 profile of CRC-32 (skips slow HD computations)
//   0x82608edb {?,?,2974,268,171,?,?,?,?,?,?,?,?}
//  the "?" entries avoid confusion about which weights were computed

// Each example shown in output is a minimum-length codeword at the HD 
//     Example: Len=2975 {0,2215,2866} (0x80000000) (Bits=4)
//  means a 2975 bit long data word with first bit (bit zero) set, bits 2215
//     and bit 2866 set gives a computed CRC result of 0x80000000 for a total
//     of four bits set in that codeword. This is an example demonstrating that
//     CRC-32 0x82608edb fails to provide HD=5 at a 2975-bit data word length. 
//     (it gives only HD=4 due to a codeword with 3 faults in data word and one 
//     fault in the check value.  Check values computed with right-shift algorithm.

// Version 2.0.0 December 2023; updated to include hash table speedup
//
// Written for g++ 9.4.0 but beware of portability problems 
//  Compile with:   g++ -O2 hdlen.cpp -o hdlen -std=c++11
// Supports up to 64-bit CRCs, but realistically good large CRCs are going
//   to be too slow to compute to get all HD values
// Note we've seen intermittent problems with higher optimization levels generating code
//       that gives incorrect results on some g++ releases, so recommend staying with -O2


#include <iostream>
#include <cstdio>
#include <sstream>
#include <stdlib.h>
#include <stdint.h>
#include <string>
#include <inttypes.h>
#include <assert.h>
#include <cstring>

using std::hex;
using std::dec;
using std::endl;
using std::cout;
using std::cerr;

// Define useful data types
typedef uint64_t Poly_t;       // CRC polynomial value, implicit +1 notation
typedef uint64_t Posn_t;       // position of a bit within a data word; data word length
typedef uint64_t BigCount_t;   // 64-bit integer for larger counts
typedef uint32_t SmallCount_t; // intentionally only 32-bits; changing this will break code
typedef uint32_t Count_t;      // used as an recursion, #bits set, or other count
typedef bool     Flag_t;       // Success flag value


static const Count_t bitsPerByte = sizeof(uint8_t)*8;
static const Count_t polySizeInBits = sizeof(Poly_t)*8;  // 64 bits in poly type

static const Count_t maxSizeHD  = polySizeInBits+8;  // HD table size
static const Posn_t unusedPosn = 0xFFFFffffFFFFffff;    // Unused HD array entry

///////////////////// General purpose helper routines ///////////////

// return number of set bits in a polynomial
static inline Count_t BitCount(Poly_t value)
{
  uint32_t count = 0;
  
  while (value != 0)
  { if (value & 1)
    { count++;
    }
    value >>= 1;
  }
  
  return(count);
}

/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
// Class HDLen: array that holds the maximum data word length for each HD
class HDLen{
 friend std::ostream& operator <<(std::ostream& hout, const HDLen& h);
 
 private:
   Poly_t hPoly = 0;
   Posn_t hDLengths[maxSizeHD+1] = {}; // initialize to 0; constructor re-initializes

 public:
   void   SetLen(Count_t weightNum, Posn_t len);  // Remember HD length
   Posn_t GetLen(Count_t weightNum);              // Get HD length
   HDLen(Poly_t poly);                      // Set currently active polynomial
   ~HDLen();                                // Destructor
};

// Constructor; remember poly and init table
HDLen::HDLen(Poly_t poly)
{ hPoly = poly;
  // Init hDLengths table to all unused entries -- done by constructor initializing to zeros
  for (Count_t i = 0; i <= maxSizeHD; i++)
  { hDLengths[i] = unusedPosn;
  }
}

// Destructor
HDLen::~HDLen()
{ // nothing to do for destruction
}

// Set a particular weight number to be a certain dataword length
inline void HDLen::SetLen(Count_t weightNum, Posn_t len)
{ hDLengths[weightNum] = len;
}

// Retrieve dataword length of a weight number.  unusedPosn if not found yet
inline Posn_t HDLen::GetLen(Count_t weightNum)
{ return(hDLengths[weightNum]);
}

// Print all computed HD lengths, putting in "?" for uncomputed lengths
std::ostream& operator <<(std::ostream& hout, const HDLen& hval)
{ auto current_flags = hout.flags();

  // print HD Lengths found surrounded by '{' '}'  
  char separator = '{';

  // print feedback polynomial
  hout << hex << "0x" << hval.hPoly << dec << " ";

  // print lengths starting at HD=3; print "?" for any uncomputed length
  Count_t weightNum = 3;
  while(weightNum <= BitCount(hval.hPoly)+1)
  { if(hval.hDLengths[weightNum] == unusedPosn)
	{ hout << separator << "?"; // ensures '{' and then ',' list separators
	} else
	{ hout << separator << hval.hDLengths[weightNum];
	}
    separator = ',';
    weightNum++;
  } // end while

  // If we ended up with a null set then print leading {
  if(separator == '{') 
  { hout << "{";
  }
  
  // finally ... end the list
  hout << "}";
  hout.flags(current_flags);
  return hout;
}

//  End of class HDLen
/////////////////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
// Class UndetectedClass: list of bit positions in the first undetected code word at HD
class UndetectedClass
{ friend std::ostream& operator <<(std::ostream& uout, const UndetectedClass& uval);

  private:
    Poly_t uPoly       = 0; // The first polynomial value undetected at this length
    Posn_t uLen        = 0; // data word length being checked for HD violation
    Poly_t uCheckValue = 0; // check value value of undetected codeword
    Posn_t posnList[maxSizeHD]; // bits set in first undetected codeword
                   // subtract 1 from these values to get 0-based bit position
  public:
         UndetectedClass(Poly_t p);    // Constructor
    void SetCheckValue(Poly_t checkValue); // Record check value
    void SetLen(Posn_t len);           // Record length for bit positioning
    void SetBitPosn(Posn_t posn, Count_t posnIndex); // Record codeword bit
    void PrintBits(Posn_t dataWordLen);             // Print example codeword
    ~UndetectedClass();
};

// Init undetected bit position data structure
UndetectedClass::UndetectedClass(Poly_t p)
{ uPoly = p;
  uCheckValue = uPoly;
  uLen = unusedPosn;
  // init all entries on list to unused
  for (Count_t posnIndex=0; posnIndex < maxSizeHD; posnIndex++)
  { posnList[posnIndex] = unusedPosn;
  }
}

UndetectedClass::~UndetectedClass()
{ // Nothing to do for destruction
}
    
// Remember the check value associated with the bits for this HD
void inline UndetectedClass::SetCheckValue(Poly_t checkValue)
{ uCheckValue = checkValue; 
} 

// Remember the length, which is needed for proper printing
// This function must be called before printing.  ASSERT used to flag this.
void inline UndetectedClass::SetLen(Posn_t len)
{ uLen = len; 
} 

// Set a particular bit position in the codeword bit list
void inline UndetectedClass::SetBitPosn(Posn_t posn, Count_t posnIndex)
{ posnList[posnIndex-1] = posn;
}

// Print out all bits in undetected codeword plus the check value
std::ostream& operator <<(std::ostream& uout, const UndetectedClass& uval)
{ auto current_flags = uout.flags();
  assert(uval.uLen != unusedPosn); // Must call SetLen first

  Count_t bitsSet = 0;
  char separator = '{';  // make separated list elements look pretty

  // print all bit positions that have been recorded
  // Note that k-bit CRC highest possible HD is k; a k+1 bit example is possible
  for(Count_t bitPosn = polySizeInBits+1;  bitPosn != 0; bitPosn--)
  { 
    // print only active entries; ignore unused entries
    if(uval.posnList[bitPosn] != unusedPosn) 
    { Posn_t thisPosn = uval.uLen - uval.posnList[bitPosn] - 1;
      uout << separator << thisPosn;
      separator = ',';
      bitsSet++;  // tally up the number of entries actually found
    }
  }

  if(separator == '{') {uout << "{";}         // print "{" if empty list
  uout << "} (0x" << hex << uval.uCheckValue << dec << ")";   // print } and check value
  
   // # bits codeword = # bits in dataword (already computed) + # bits in check value 
  bitsSet += BitCount(uval.uCheckValue); 
  uout << " (Bits=" << bitsSet << ")";
  uout.flags(current_flags);
  return uout;
}

// End Class UndetectedClass
/////////////////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
// Class LFSR: build and manage an LFSR lookup capability
//   Provides way to look up the nth LFSR value from a rolling "1" bit start point
class LFSR{
  private:         
     Poly_t myPoly = 0;               // Current CRC polynomial
     
  public:
     inline Poly_t RollByOne(Poly_t val);    // Roll right-shift LFSR algorithm one bit
    
     LFSR(Poly_t newPolynomial);                     // Constructor
     ~LFSR();                                        // Destructor
     LFSR(const LFSR&);			                         // Shallow copy
	   LFSR& operator =(const LFSR&);                  // Shallow copy
};

LFSR& LFSR::operator =(const LFSR&)  = default;
LFSR::LFSR(const LFSR&) = default;

// LFSR Constructor.  
LFSR::LFSR(Poly_t newPolynomial)
{
  myPoly = newPolynomial;
}

// Destructor
LFSR::~LFSR()
{
}

// Roll CRC 1 bit using shift and conditional XOR of the polynomial value
inline Poly_t LFSR::RollByOne(Poly_t val)
{ if(val & 1)
  { val >>= 1;
    val ^= myPoly;
  } else
  { val >>= 1;
  } 
  return(val);
}

// End Class HashTable
/////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
///////////////////// Polynomial Class //////////////////////////////////
class CRCpoly{
  private:         
    Poly_t  polynomial = 0;   // implicit +1 polynomial representation
    
    HDLen * HDArray = NULL;              // HDLen array for this poly
    UndetectedClass * Undetected = NULL; // Undetected bit array for this poly
    LFSR * myLFSR = NULL;                // LFSR capability

    Posn_t FindHD(Count_t hdGoal);       // Base case for iteration
    Flag_t   FindHDRecurse(              // Dive deep to find HD general case
                Poly_t accum, Posn_t posn, Count_t hdBitsLeft);
    Posn_t FindHDOther(Count_t hdGoal);               // Optimized helper for HD>=4 case
    inline Flag_t CheckAccum(            // Helper to check if accum value violates HD
                Poly_t accum, Posn_t posn, Count_t hdBitsLeft);

  public:
    inline Poly_t Poly();                 // Retrieve binary polynomial value
    
    void   ProcessCRC(                    // Iterate across all HDs in search                
             Count_t startHD, Count_t maxHD);
				   
    ~CRCpoly();                          // Destructor; release resources
    CRCpoly(Poly_t crcPoly);             // Constructor
    CRCpoly(const CRCpoly&);			       // Shallow copy
	  CRCpoly& operator =(const CRCpoly&); // Shallow copy
};

// Shallow copy is OK for this class
 CRCpoly& CRCpoly::operator =(const CRCpoly&)  = default;
 CRCpoly::CRCpoly(const CRCpoly&) = default;

// Constructor. Cache results of characterization of poly and set up lists
CRCpoly::CRCpoly(Poly_t poly)
{ 
  polynomial = poly;
  
  // HDArray used to record HD lengths
  HDArray = new HDLen(poly);
  // Undetected array used to record bits in example undetected codeword
  Undetected = new UndetectedClass(poly);
    
  myLFSR = new LFSR(polynomial);
}

// Destructor -- release HDArray and Undetected to avoid memory leaks
CRCpoly::~CRCpoly()
{ delete HDArray;
  delete Undetected;
  delete myLFSR;
}

// Retrieve binary polynomial value
inline Poly_t CRCpoly::Poly()
{ return(polynomial);
}

// Helper to determine if residual number of bits in a computed check value
//  is small enough to violate the HD.
inline Flag_t CRCpoly::CheckAccum(Poly_t accum, 
                    Posn_t posn, Count_t hdBitsLeft)
{   
  // See if Accum value has fewer bits than required to hit or exceed the HD
  //   Note: the "-1" is because accum implicitly has one additional bit set
  //     This is done to make the calling code cleaner.
  Flag_t successFlag = (BitCount(accum) < hdBitsLeft-1);

  if (successFlag)
  { 
    // too few bits are set in check value, so this codeword fails to provide HD
    Undetected->SetCheckValue(accum);   // Record check value and this bit position
    Undetected->SetBitPosn(posn, hdBitsLeft);          
  }

  return(successFlag);  // if true, have found the length for this HD
}

// Recursively find whether this polynomial at this length provides the specified HD 
// Dives into the middle of a computation with accumulator representing
//   the aggregate contribution to final check value value from previous bits
// Length starts from bit that is the last bit before the check value and works
//   toward longer word lengths.  The data word bit at the maxPosn position
//   has its bit set from the calling level of recursion. maxPosn must be 1 or greater   
Flag_t CRCpoly::FindHDRecurse(
                         Poly_t accum, Posn_t maxPosn, Count_t hdBitsLeft)
{ 
  // cerr << "recurse hdBitsLeft=" << hdBitsLeft << endl;
  Posn_t posn = 0;  
  Poly_t rollingValue = this->polynomial; // position zero LFSR value

  // Posn 0 is special case -- no possibility of recursion because no bits left
  //   But need to check this in case FCS has fewer bits than HD
  Poly_t newAccum = rollingValue ^ accum;
  Flag_t successFlag = CheckAccum(newAccum, posn, hdBitsLeft);
  
  posn++; // Try the next bit at this level of recursion 

  // Try bit positions 1 and higher while have not violated HD
  while ((posn < maxPosn) && (!successFlag))
  { 
    // Add one more bit fault at this position
    rollingValue = myLFSR->RollByOne(rollingValue);
    newAccum = rollingValue ^ accum;

    // See if this check value is worse than HD with just this one bit set
    if(CheckAccum(newAccum, posn, hdBitsLeft))  
    { successFlag = true;
    }
    else
    { // recurse to add contribution of the next bit in the dataword
      //  we know already that posn is 1 or higher, so don't need to check for posn==0  
      if(hdBitsLeft > 1) { successFlag = FindHDRecurse(newAccum, posn, hdBitsLeft-1);  }
    }
    
    // Don't move posn++ to start of loop - it upsets the first loop bounds logic
    posn++; // Try the next bit at this level of recursion 
  }  // end while
  
  // found an HD violation, so record it here
  if (successFlag)
  { // "posn-1" compensates for posn++ at end of loop instead of start
    Undetected->SetBitPosn(posn-1, hdBitsLeft);  
  }
  
  return(successFlag); // Returns false if HD violation
}

// Find HD for all HDs
Posn_t CRCpoly::FindHDOther(Count_t hdGoal)
{
  Poly_t accum = Poly();  // Accumulator walking first bit error                    
  Posn_t posn = 0;

  delete Undetected;
  Undetected = new UndetectedClass(polynomial);
    
  // Set up the first bit fault that defines the effective length of the data word
  // Loop stops when the HD is found
  //    The while condition is for  1 bit fault in data word + check value >= HD
  // By design, this setup skips recursion on position 0 because no room left in data word then
  while(!CheckAccum(accum, posn, hdGoal)) // 1 fault bit in data word + bits in check value
  {  
    // advance the first bit further away from the check value by 1 bit
    posn++; 
    accum = myLFSR->RollByOne(accum); 

    // Recurse to see if there is a HD violation at this length
    //    The recursion sets additional bit faults in the data word  
    if (hdGoal > 1) { if(FindHDRecurse(accum, posn, hdGoal-1)) { break; } }
  }

  // found the first error that exceeds HD
  Undetected->SetBitPosn(posn, hdGoal);
  return(posn);
}

// Outer loop to find longest dataword length at a particular HD
Posn_t CRCpoly::FindHD(Count_t hdGoal)
{ 
  Posn_t posn = 0;
 
  posn = FindHDOther(hdGoal);  

  // Record the final length of what we found
  HDArray->SetLen(hdGoal,posn);     
  Undetected->SetLen(posn+1);  
 
  cout << "# 0x" << hex << Poly() << dec << "  HD=" << hdGoal;
  if(posn == 0)  // HD=2 on one-bit data word
  { cout << "  NONE  ";
  } else  // have found a length that supports this HD
  { cout << "  len=" << posn << "  "; 
  }
  
  // Print the example showing next longer dataword violates HD
  cout << "Example: Len=" << posn+1 << " ";

  // Exit loop when first bit has found what we are looking for
  cout << *Undetected << endl;
  return(posn);
}

// Print selected range of HDs for a polynomial
void CRCpoly::ProcessCRC(Count_t startHD, Count_t maxHD)
{
   // All CRCs have HD=2 at infinite length, so clamp starting HD at 3
   if(startHD < 3)  { startHD = 3;}
   
   // All CRCs violate HD= # active bits in polynomial+1  at length=1 (position 0)
   //   So clamp max HD at that number +1 to show the "NONE" entry
   // do something reasonable for invalid maxHD input
   if(   (maxHD > BitCount(polynomial)+2)
      || (maxHD < startHD) 
     )
   { maxHD = BitCount(polynomial)+2; 
   }

   // Status to stderr to track progress when running a batch
   std::cerr << "Poly=0x" << hex << polynomial << dec;
   std::cerr << " startHD=" << startHD << " maxHD=" << maxHD << endl;

   // Find data word lengths for requested HD range
   for (Count_t currentHD = startHD; currentHD <= maxHD; currentHD++)
   { FindHD(currentHD);    // all other HDs
   } 
   
   // Print summary findings as last item
   cout << *HDArray << endl;
   
}


//////////////////////////// Main //////////////////////////////////////////////

int main(int argc, char **argv)
{
  Poly_t polyValue = 0;      // Working CRC polynomial
  Count_t startHD = 0;       // Requested start HD range, if any
  Count_t maxHD = 0;         // Requested max HD range, if any

  Flag_t useStdin = false;   // Are we getting inputs from STDIN?
  std::istringstream cmd;    // Command line argument stream
  Flag_t failFlag = false;   // Command line parsing failure if true

  switch (argc)
  { //  CMD:   ./hdlen   <polyfile.txt
    case 1: // no command line inputs; full HD from stdin
       useStdin = true;
    break;

    //  CMD:   ./hdlen 0x82608edb      
    case 2: // poly as only command line input
       useStdin = false;
       cmd.clear(); cmd.str(argv[1]); if(!(cmd >> hex >> polyValue )) {failFlag =true;}
    break;


    //  CMD:   ./hdlen 5 7  <polyfile.txt     
    case 3: // minHD and maxHD as command line inputs
       useStdin = true;
       cmd.clear(); cmd.str(argv[1]); if(!(cmd >> dec >> startHD   )) {failFlag =true;}
       cmd.clear(); cmd.str(argv[2]); if(!(cmd >> dec >> maxHD     )) {failFlag =true;}
    break;

    //  CMD:   ./hdlen 0x82608edb 5 7      
    case 4: // poly, minHD and maxHD as command line inputs
       useStdin = false;
       cmd.clear(); cmd.str(argv[1]); if(!(cmd >> hex >> polyValue )) {failFlag =true;}
       cmd.clear(); cmd.str(argv[2]); if(!(cmd >> dec >> startHD   )) {failFlag =true;}
       cmd.clear(); cmd.str(argv[3]); if(!(cmd >> dec >> maxHD     )) {failFlag =true;}
    break;

    default: // invalid number of command line parameters
       failFlag = true;
    break;      
  } // end of switch

  if(failFlag)
  { std::cerr << "Usage: " << argv[0]  
            << " [poly] | [StartHD MaxHD] | [poly StartHD MaxHD]" << endl;
  } else
  { do //  do ... while to ensure always process at least one poly if available 
    { 
      // Get next polynomial to analyze
      if(useStdin)
      { std::cin >> hex >> polyValue >> dec;  
        if (std::cin.fail()) { break; }
      }

      // Perform complete computation for one polynomial
      CRCpoly * Poly = new CRCpoly(polyValue);

      Poly->ProcessCRC(startHD, maxHD);  
      delete Poly;
      
    } while ((!feof(stdin)) && useStdin);   // end while
  }

  return(failFlag);
}